import logging
from pathlib import Path
from typing import Any, Union

import numpy as np
import rasterio

from segmentmytif.logging_config import log_array

logger = logging.getLogger(__name__)


def read_geotiff(input_path: Path) -> (np.ndarray, Any):
    with rasterio.open(input_path) as src:
        data = src.read()
        profile = src.profile
    log_array(data, logger, array_name=input_path)
    return data, profile


def save_tiff(data: np.ndarray, output_path: Union[Path, str], profile) -> None:
    profile.update(count=data.shape[0])  # set number of channels
    profile.update(compress=None)
    with rasterio.open(str(output_path), 'w', **profile) as dst:
        dst.write(data)



import json
from pathlib import Path

import torch
from torchinfo import summary

from segmentmytif.utils.models import UNet
from tests.utils import TEST_DATA_FOLDER


class TestUNet:
    def test_initialize_unet_with_valid_parameters(self):
        """Initialize UNet with valid in_channels and num_classes"""
        in_channels = 3
        num_classes = 19

        model = UNet(in_channels, num_classes)

        assert isinstance(model, UNet)
        assert model.out.out_channels == num_classes

    def test_forward_with_minimum_valid_dimensions(self):
        """Handle input tensor with minimum valid dimensions"""
        in_channels = 3
        num_classes = 19
        width = height = 64
        input_tensor = torch.randn(1, in_channels, width, height)  # Minimum size for U-Net to work

        model = UNet(in_channels, num_classes)
        output = model(input_tensor)

        assert output.shape == (1, num_classes, width, height)

    def test_summary(self):
        """Summary of model is exactly as tested."""
        in_channels = 3
        num_classes = 19
        width = height = 64
        input_tensor = torch.randn(1, in_channels, width, height)  # Minimum size for U-Net to work
        expected_sum = load_summary('test_model_summary.json')

        model = UNet(in_channels, num_classes)
        sum = get_summary(model, input_tensor)

        assert sum == expected_sum

    def test_summary_model_scaled_down(self):
        """Summary of model is exactly as tested."""
        model_scale = 0.5
        in_channels = 3
        num_classes = 19
        width = height = 64
        input_tensor = torch.randn(1, in_channels, width, height)  # Minimum size for U-Net to work
        expected_sum = load_summary('test_model_summary_half.json')

        model = UNet(in_channels, num_classes, model_scale=model_scale)
        sum = get_summary(model, input_tensor)

        assert sum == expected_sum


def load_summary(file_name):
    expected_sum = (TEST_DATA_FOLDER / file_name).read_text(encoding='utf-8')
    return normalize_model_summary(expected_sum)


def get_summary(model, input_tensor):
    return normalize_model_summary(summary(model, input_data=input_tensor, verbose=0).__repr__())


def normalize_model_summary(text):
    return (text
            .replace("=", "")
            .replace("(G)", "(Units.GIGABYTES)")
            .replace("(M)", "(Units.MEGABYTES)")
            )



import shutil
from pathlib import Path

import dask.array as da
import numpy as np
import pytest
import rasterio
import itertools

from segmentmytif.features import FeatureType, get_features_path
from segmentmytif.main import read_input_and_labels_and_save_predictions, prepare_training_data
from segmentmytif.utils.io import save_tiff
from .utils import TEST_DATA_FOLDER

# Test integration with a cmbination of inputs
test_images_and_labels = [
    ("test_image.tif", "test_image_labels_positive.gpkg", "test_image_labels_negative.gpkg"),
    ("test_image_512x512.tif", "test_image_labels_positive_512x512.gpkg", "test_image_labels_negative_512x512.gpkg"),
]
feature_types = [FeatureType.IDENTITY, FeatureType.FLAIR]
compute_modes = ["normal", "parallel", "safe"]
combinations = list(itertools.product(test_images_and_labels, feature_types, compute_modes))


@pytest.mark.parametrize(
    ("test_images_and_labels", "feature_type", "compute_modes"),
    combinations,
)
def test_integration(tmpdir, test_images_and_labels, feature_type, compute_modes):
    input_path = copy_file_and_get_new_path(test_images_and_labels[0], tmpdir)
    labels_pos_path = copy_file_and_get_new_path(test_images_and_labels[1], tmpdir)
    labels_neg_path = copy_file_and_get_new_path(test_images_and_labels[2], tmpdir)
    predictions_path = Path(tmpdir) / f"{test_images_and_labels[0]}_predictions_{str(feature_type)}.tif"

    read_input_and_labels_and_save_predictions(
        input_path, labels_pos_path, labels_neg_path, predictions_path, feature_type=feature_type, model_scale=0.125
    )  # scale down feature-extraction-model for testing

    assert predictions_path.exists()


def copy_file_and_get_new_path(test_image, tmpdir):
    input_path = Path(tmpdir) / test_image
    shutil.copy(TEST_DATA_FOLDER / test_image, input_path)
    return input_path


@pytest.mark.parametrize(
    "input_path, feature_type, expected_path",
    [
        ("input.tiff", FeatureType.FLAIR, "input_FLAIR.tiff"),
        ("../path/to/input.tiff", FeatureType.FLAIR, "../path/to/input_FLAIR.tiff"),
        ("../path/to/input.tiff", FeatureType.IDENTITY, "../path/to/input.tiff"),
    ],
)
def test_get_features_path(input_path, feature_type, expected_path):
    features_path = get_features_path(Path(input_path), feature_type)
    assert features_path == Path(expected_path)


def test_save(tmpdir):
    predictions_path = Path(tmpdir) / "test_image_predictions.tif"
    width = 100
    data = np.zeros((3, width, width))
    profile = {"width": width, "height": width, "dtype": rasterio.uint8}

    save_tiff(data, predictions_path, profile=profile)

    assert predictions_path.exists()


@pytest.mark.parametrize("array_type", ["numpy", "dask"])
def test_prepare_training_data(array_type):
    random = np.random.default_rng(0)
    length = 200
    random_data = random.integers(low=0, high=256, size=(5, length, length))
    labels = random.choice([0, 1, 2], size=(1, length, length), replace=True)
    if array_type == "numpy":
        input_data = random_data
    elif array_type == "dask":
        input_data = da.from_array(random_data)

    prepare_training_data(input_data, labels)
